#ifndef TREENODEMODEL_H
#define TREENODEMODEL_H

#include "gmc-common_global.h"

#include <QAbstractItemModel>
#include <QIdentityProxyModel>
#include <QSortFilterProxyModel>

#include <QSet>

#include "treenodemodelextension.h"

class GMC_COMMON_SHARED_EXPORT TreeNodeModel
: public QAbstractItemModel
, virtual public TreeNodeModelExtension
{

    Q_OBJECT

public:

    TreeNodeModel( QObject * /*parent*/ = nullptr);
    TreeNodeModel( TreeNode * /*rootNode*/, QObject * /*parent*/ = nullptr);
    virtual ~TreeNodeModel();

    void setRootNode( TreeNode * /*rootNode*/ );

    QVariant data( const QModelIndex & /*index*/
                   , int /*role*/ = Qt::DisplayRole ) const override;
    bool setData( const QModelIndex & /*index*/
                  , const QVariant & /*value*/
                  , int /*role*/ = Qt::EditRole ) override;
    Qt::ItemFlags flags(const QModelIndex & /*index*/ ) const override;
    QModelIndex index( int /*row*/
                       , int /*column*/
                       , const QModelIndex & /*parentIndex*/ = QModelIndex() ) const override;

    QModelIndex parent( const QModelIndex & /*index*/ ) const override;
    int rowCount( const QModelIndex & /*index*/ = QModelIndex() ) const override;
    int columnCount( const QModelIndex & /*index*/ = QModelIndex() ) const override;
    QVariant headerData( int /*section*/
                               , Qt::Orientation /*orientation*/
                               , int /*role*/ = Qt::DisplayRole ) const override;

    bool canFetchMore(const QModelIndex & /*parent*/ ) const override;
    void fetchMore(const QModelIndex & /*parent*/) override;
    bool hasChildren(const QModelIndex & /*parent*/ ) const override;

    TreeNode *rootNode() const override;

    TreeNode * toTreeNode( const QModelIndex & /*index*/ ) const override;
    TreeNode * toTreeNode( const QItemSelectionRange & /*selectionRange*/ ) const override;
    TreeNode * toTreeNode( const QItemSelection & /*selection*/ ) const override;

    QModelIndex treeNodeToIndex( TreeNode * /*node*/ ) const override;

    const QList< TreeNode * > &extraRootChildNodes() const;
    void setExtraRootChildNodes( const QList<TreeNode *> &/*extraRootChildNodes*/ );
    void deleteExtraRootChildNodes();

protected:

    void connectVertexNode( TreeNode * /*node*/ );
    void disconnectVertexNode( TreeNode * /*node*/ );

public slots:

    void nodesAboutToBeReset();
    void nodesReset();

protected slots:

    void nodesAboutToBeInserted( TreeNode * /*parent*/, int /*start*/, int /*end*/ );
    void nodesAboutToBeRemoved( TreeNode * /*parent*/, int /*start*/, int /*end*/ );
    void nodesInserted();
    void nodesRemoved();
    void nodeDataChanged( TreeNode * /*node*/ );
    void nodeChildsDataChanged( TreeNode * /*node*/ );

protected:

    TreeNode *_rootNode;
    QList< TreeNode * > _extraRootChildNodes;

};

//******************************************************************************

class GMC_COMMON_SHARED_EXPORT TreeNodeSortFilterProxyModel
: public QSortFilterProxyModel
, virtual public TreeNodeModelExtension
{

    Q_OBJECT

public:

    enum SpecialFilterType { UndefinedFilterType             = 0x0
                           , InvisibleFilterType             = 0x1
                           , SelectableFilterType            = 0x2 };


    TreeNodeSortFilterProxyModel( int /*specialFilters*/ = TreeNodeSortFilterProxyModel::UndefinedFilterType
                                , QObject * /*parent*/ = nullptr );
    TreeNodeSortFilterProxyModel( QSet< int > /*acceptColumns*/
                                , int /*specialFilters*/ = TreeNodeSortFilterProxyModel::UndefinedFilterType
                                , QObject * /*parent*/ = nullptr );
    virtual ~TreeNodeSortFilterProxyModel();

    const QSet<int> &acceptColumns() const;
    void setAcceptColumns( const QSet<int> &acceptColumns );

    TreeNode * rootNode() const override;

    TreeNode * toTreeNode( const QModelIndex & /*index*/ ) const override;
    TreeNode * toTreeNode( const QItemSelectionRange & /*selectionRange*/ ) const override;
    TreeNode * toTreeNode( const QItemSelection & /*selection*/ ) const override;

    QModelIndex treeNodeToIndex( TreeNode * /*treeNode*/ ) const override;

protected:

    bool filterAcceptsColumn( int /*srcColumn*/, const QModelIndex & /*srcParentIndex*/ ) const override;
    bool filterAcceptsRow( int /*srcRow*/, const QModelIndex & /*srcParentIndex*/ ) const override;

protected:

    int _specialFilters;
    QSet< int > _acceptColumns;

};

//******************************************************************************

class GMC_COMMON_SHARED_EXPORT FilterTreeNodeSortFilterProxyModel : public TreeNodeSortFilterProxyModel
{
    Q_OBJECT

public:

    FilterTreeNodeSortFilterProxyModel( int /*specialFilters*/ = TreeNodeSortFilterProxyModel::UndefinedFilterType
                                      , QObject * /*parent*/ = nullptr );
    FilterTreeNodeSortFilterProxyModel( QSet< int > /*acceptsColumn*/
                                      , int /*specialFilters*/ = TreeNodeSortFilterProxyModel::UndefinedFilterType
                                      , QObject * /*parent*/ = nullptr );
    virtual ~FilterTreeNodeSortFilterProxyModel();

protected:

    bool filterAcceptsRow( int /*srcRow*/, const QModelIndex & /*srcParentIndex*/ ) const override;

};

//******************************************************************************

class GMC_COMMON_SHARED_EXPORT ChoiceTreeNodeProxyModel
: public QIdentityProxyModel
, virtual public TreeNodeModelExtension
{

    Q_OBJECT

signals:

    void stateChanged();

public:

    ChoiceTreeNodeProxyModel( int /*nameColumn*/
                            , int /*interceptChoiceRole*/
                            , int /*interceptBlockRole*/
                            , QObject * /*parent*/ = nullptr );
    virtual ~ChoiceTreeNodeProxyModel();

    int nameColumn() const;
    int choiceRole() const;
    int blockRole() const;

    QVariant data(const QModelIndex & /*index*/, int /*role*/ = Qt::DisplayRole) const override;
    bool setData(const QModelIndex & /*index*/, const QVariant & /*value*/, int /*role*/ = Qt::EditRole ) override;
    Qt::ItemFlags flags(const QModelIndex & /*index*/ ) const override;

    void setCheckedBlockedNames( const QSet< QString > /*checkedNames*/
                               , const QSet< QString > /*blockedNames*/ );
    void setCheckedNames( const QSet< QString > /*checkedNames*/ );
    void setBlockedNames( const QSet< QString > /*blockedNames*/ );

    const QSet<QString> &checkedNames() const;
    const QSet<QString> &blockedNames() const;

    const QSet<QString> &addNames() const;
    const QSet<QString> &removeNames() const;
    void clearAddRemoveNames();


    TreeNode * rootNode() const override;

    TreeNode * toTreeNode( const QModelIndex & /*index*/ ) const override;
    TreeNode * toTreeNode( const QItemSelectionRange & /*selectionRange*/ ) const override;
    TreeNode * toTreeNode( const QItemSelection & /*selection*/ ) const override;

    QModelIndex treeNodeToIndex( TreeNode * /*treeNode*/ ) const override;

protected:

    int _nameColumn;
    int _choiceRole;
    int _blockRole;

    QSet< QString > _checkedNames;
    QSet< QString > _blockedNames;

    QSet< QString > _addNames;
    QSet< QString > _removeNames;

};

//******************************************************************************

class GMC_COMMON_SHARED_EXPORT ChoiceTreeNodeSortFilterProxyModel
: public TreeNodeSortFilterProxyModel
{

    Q_OBJECT

public:

    ChoiceTreeNodeSortFilterProxyModel( bool /*acceptValue*/
                                      , QObject * /*parent*/ = nullptr );

    virtual ~ChoiceTreeNodeSortFilterProxyModel();

protected:

    bool filterAcceptsRow( int /*srcRow*/, const QModelIndex & /*srcParentIndex*/ ) const override;

protected:

    bool _acceptValue;

};

//******************************************************************************

class GMC_COMMON_SHARED_EXPORT ChoiceFilterTreeNodeSortFilterProxyModel
: public FilterTreeNodeSortFilterProxyModel
{

    Q_OBJECT

public:

    ChoiceFilterTreeNodeSortFilterProxyModel( bool /*acceptValue*/
                                            , QObject * /*parent*/ = nullptr );

    virtual ~ChoiceFilterTreeNodeSortFilterProxyModel();

protected:

    bool filterAcceptsRow( int /*srcRow*/, const QModelIndex & /*srcParentIndex*/ ) const override;

protected:

    bool _acceptValue;

};

#endif // TREENODEMODEL_H
